/*
 * Copyright 2017-2021 Philip Rebohle
 * Copyright 2024 Collabora Ltd.
 * SPDX-License-Identifier: Zlib
 *
 * This is a cut-down version of the header from
 * https://github.com/misyltoad/dxvk-native/tree/master/tests
 * with only the parts strictly required by tests/native/
 */

#pragma once

#include <array>
#include <iostream>

#include "../../../src/util/com/com_pointer.h"
#include "../../../src/util/util_error.h"
